"""Generated wrapper for ERC721Token Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for ERC721Token below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        ERC721TokenValidator,
    )
except ImportError:

    class ERC721TokenValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


class GetApprovedMethod(ContractMethod):
    """Various interfaces to the getApproved method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _token_id: int):
        """Validate the inputs to the getApproved method."""
        self.validator.assert_valid(
            method_name="getApproved",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        return _token_id

    def call(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> str:
        """Execute underlying contract method via eth_call.

        Throws if `_tokenId` is not a valid NFT.

        :param _tokenId: The NFT to find the approved address for
        :param tx_params: transaction parameters
        :returns: The approved address for this NFT, or the zero address if
            there is none
        """
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).call(tx_params.as_dict())

    def send_transaction(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Throws if `_tokenId` is not a valid NFT.

        :param _tokenId: The NFT to find the approved address for
        :param tx_params: transaction parameters
        :returns: The approved address for this NFT, or the zero address if
            there is none
        """
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).transact(tx_params.as_dict())

    def estimate_gas(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).buildTransaction(
            tx_params.as_dict()
        )


class ApproveMethod(ContractMethod):
    """Various interfaces to the approve method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _approved: str, _token_id: int):
        """Validate the inputs to the approve method."""
        self.validator.assert_valid(
            method_name="approve",
            parameter_name="_approved",
            argument_value=_approved,
        )
        _approved = self.validate_and_checksum_address(_approved)
        self.validator.assert_valid(
            method_name="approve",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        return (_approved, _token_id)

    def call(
        self,
        _approved: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        The zero address indicates there is no approved address. Throws unless
        `msg.sender` is the current NFT owner, or an authorized operator of the
        current owner.

        :param _approved: The new approved NFT controller
        :param _tokenId: The NFT to approve
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_approved, _token_id) = self.validate_and_normalize_inputs(
            _approved, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_approved, _token_id).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        _approved: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        The zero address indicates there is no approved address. Throws unless
        `msg.sender` is the current NFT owner, or an authorized operator of the
        current owner.

        :param _approved: The new approved NFT controller
        :param _tokenId: The NFT to approve
        :param tx_params: transaction parameters
        """
        (_approved, _token_id) = self.validate_and_normalize_inputs(
            _approved, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_approved, _token_id).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        _approved: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (_approved, _token_id) = self.validate_and_normalize_inputs(
            _approved, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_approved, _token_id).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self,
        _approved: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_approved, _token_id) = self.validate_and_normalize_inputs(
            _approved, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_approved, _token_id).buildTransaction(
            tx_params.as_dict()
        )


class TransferFromMethod(ContractMethod):
    """Various interfaces to the transferFrom method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(
        self, _from: str, _to: str, _token_id: int
    ):
        """Validate the inputs to the transferFrom method."""
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="_from",
            argument_value=_from,
        )
        _from = self.validate_and_checksum_address(_from)
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="_to",
            argument_value=_to,
        )
        _to = self.validate_and_checksum_address(_to)
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        return (_from, _to, _token_id)

    def call(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Throws unless `msg.sender` is the current owner, an authorized
        operator, or the approved address for this NFT. Throws if `_from` is
        not the current owner. Throws if `_to` is the zero address. Throws if
        `_tokenId` is not a valid NFT.

        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Throws unless `msg.sender` is the current owner, an authorized
        operator, or the approved address for this NFT. Throws if `_from` is
        not the current owner. Throws if `_to` is the zero address. Throws if
        `_tokenId` is not a valid NFT.

        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        """
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).buildTransaction(
            tx_params.as_dict()
        )


class SafeTransferFrom1Method(ContractMethod):
    """Various interfaces to the safeTransferFrom method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(
        self, _from: str, _to: str, _token_id: int
    ):
        """Validate the inputs to the safeTransferFrom method."""
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_from",
            argument_value=_from,
        )
        _from = self.validate_and_checksum_address(_from)
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_to",
            argument_value=_to,
        )
        _to = self.validate_and_checksum_address(_to)
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        return (_from, _to, _token_id)

    def call(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        This works identically to the other function with an extra data
        parameter, except this function just sets data to "".

        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        This works identically to the other function with an extra data
        parameter, except this function just sets data to "".

        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        """
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_from, _to, _token_id) = self.validate_and_normalize_inputs(
            _from, _to, _token_id
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id).buildTransaction(
            tx_params.as_dict()
        )


class OwnerOfMethod(ContractMethod):
    """Various interfaces to the ownerOf method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _token_id: int):
        """Validate the inputs to the ownerOf method."""
        self.validator.assert_valid(
            method_name="ownerOf",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        return _token_id

    def call(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> str:
        """Execute underlying contract method via eth_call.

        NFTs assigned to zero address are considered invalid, and queries about
        them do throw.

        :param _tokenId: The identifier for an NFT
        :param tx_params: transaction parameters
        :returns: The address of the owner of the NFT
        """
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).call(tx_params.as_dict())

    def send_transaction(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        NFTs assigned to zero address are considered invalid, and queries about
        them do throw.

        :param _tokenId: The identifier for an NFT
        :param tx_params: transaction parameters
        :returns: The address of the owner of the NFT
        """
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).transact(tx_params.as_dict())

    def estimate_gas(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, _token_id: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_token_id) = self.validate_and_normalize_inputs(_token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_token_id).buildTransaction(
            tx_params.as_dict()
        )


class BalanceOfMethod(ContractMethod):
    """Various interfaces to the balanceOf method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _owner: str):
        """Validate the inputs to the balanceOf method."""
        self.validator.assert_valid(
            method_name="balanceOf",
            parameter_name="_owner",
            argument_value=_owner,
        )
        _owner = self.validate_and_checksum_address(_owner)
        return _owner

    def call(self, _owner: str, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        NFTs assigned to the zero address are considered invalid, and this
        function throws for queries about the zero address.

        :param _owner: An address for whom to query the balance
        :param tx_params: transaction parameters
        :returns: The number of NFTs owned by `_owner`, possibly zero
        """
        (_owner) = self.validate_and_normalize_inputs(_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner).call(tx_params.as_dict())

    def send_transaction(
        self, _owner: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        NFTs assigned to the zero address are considered invalid, and this
        function throws for queries about the zero address.

        :param _owner: An address for whom to query the balance
        :param tx_params: transaction parameters
        :returns: The number of NFTs owned by `_owner`, possibly zero
        """
        (_owner) = self.validate_and_normalize_inputs(_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner).transact(tx_params.as_dict())

    def estimate_gas(
        self, _owner: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (_owner) = self.validate_and_normalize_inputs(_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner).estimateGas(tx_params.as_dict())

    def build_transaction(
        self, _owner: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_owner) = self.validate_and_normalize_inputs(_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner).buildTransaction(
            tx_params.as_dict()
        )


class SetApprovalForAllMethod(ContractMethod):
    """Various interfaces to the setApprovalForAll method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _operator: str, _approved: bool):
        """Validate the inputs to the setApprovalForAll method."""
        self.validator.assert_valid(
            method_name="setApprovalForAll",
            parameter_name="_operator",
            argument_value=_operator,
        )
        _operator = self.validate_and_checksum_address(_operator)
        self.validator.assert_valid(
            method_name="setApprovalForAll",
            parameter_name="_approved",
            argument_value=_approved,
        )
        return (_operator, _approved)

    def call(
        self,
        _operator: str,
        _approved: bool,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Emits the ApprovalForAll event. The contract MUST allow multiple
        operators per owner.

        :param _approved: True if the operator is approved, false to revoke
            approval
        :param _operator: Address to add to the set of authorized operators
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_operator, _approved) = self.validate_and_normalize_inputs(
            _operator, _approved
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_operator, _approved).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        _operator: str,
        _approved: bool,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Emits the ApprovalForAll event. The contract MUST allow multiple
        operators per owner.

        :param _approved: True if the operator is approved, false to revoke
            approval
        :param _operator: Address to add to the set of authorized operators
        :param tx_params: transaction parameters
        """
        (_operator, _approved) = self.validate_and_normalize_inputs(
            _operator, _approved
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_operator, _approved).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        _operator: str,
        _approved: bool,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (_operator, _approved) = self.validate_and_normalize_inputs(
            _operator, _approved
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_operator, _approved).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self,
        _operator: str,
        _approved: bool,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_operator, _approved) = self.validate_and_normalize_inputs(
            _operator, _approved
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_operator, _approved).buildTransaction(
            tx_params.as_dict()
        )


class SafeTransferFrom2Method(ContractMethod):
    """Various interfaces to the safeTransferFrom method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(
        self, _from: str, _to: str, _token_id: int, _data: bytes
    ):
        """Validate the inputs to the safeTransferFrom method."""
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_from",
            argument_value=_from,
        )
        _from = self.validate_and_checksum_address(_from)
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_to",
            argument_value=_to,
        )
        _to = self.validate_and_checksum_address(_to)
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_tokenId",
            argument_value=_token_id,
        )
        # safeguard against fractional inputs
        _token_id = int(_token_id)
        self.validator.assert_valid(
            method_name="safeTransferFrom",
            parameter_name="_data",
            argument_value=_data,
        )
        _data = bytes.fromhex(_data.decode("utf-8"))
        return (_from, _to, _token_id, _data)

    def call(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        _data: bytes,
        tx_params: Optional[TxParams] = None,
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        Throws unless `msg.sender` is the current owner, an authorized
        operator, or the approved address for this NFT. Throws if `_from` is
        not the current owner. Throws if `_to` is the zero address. Throws if
        `_tokenId` is not a valid NFT. When transfer is complete, this function
        checks if `_to` is a smart contract (code size > 0). If so, it calls
        `onERC721Received` on `_to` and throws if the return value is not
        `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`.

        :param _data: Additional data with no specified format, sent in call to
            `_to`
        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_from, _to, _token_id, _data) = self.validate_and_normalize_inputs(
            _from, _to, _token_id, _data
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id, _data).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        _data: bytes,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        Throws unless `msg.sender` is the current owner, an authorized
        operator, or the approved address for this NFT. Throws if `_from` is
        not the current owner. Throws if `_to` is the zero address. Throws if
        `_tokenId` is not a valid NFT. When transfer is complete, this function
        checks if `_to` is a smart contract (code size > 0). If so, it calls
        `onERC721Received` on `_to` and throws if the return value is not
        `bytes4(keccak256("onERC721Received(address,address,uint256,bytes)"))`.

        :param _data: Additional data with no specified format, sent in call to
            `_to`
        :param _from: The current owner of the NFT
        :param _to: The new owner
        :param _tokenId: The NFT to transfer
        :param tx_params: transaction parameters
        """
        (_from, _to, _token_id, _data) = self.validate_and_normalize_inputs(
            _from, _to, _token_id, _data
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, _to, _token_id, _data).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        _data: bytes,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (_from, _to, _token_id, _data) = self.validate_and_normalize_inputs(
            _from, _to, _token_id, _data
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(
            _from, _to, _token_id, _data
        ).estimateGas(tx_params.as_dict())

    def build_transaction(
        self,
        _from: str,
        _to: str,
        _token_id: int,
        _data: bytes,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_from, _to, _token_id, _data) = self.validate_and_normalize_inputs(
            _from, _to, _token_id, _data
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(
            _from, _to, _token_id, _data
        ).buildTransaction(tx_params.as_dict())


class IsApprovedForAllMethod(ContractMethod):
    """Various interfaces to the isApprovedForAll method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _owner: str, _operator: str):
        """Validate the inputs to the isApprovedForAll method."""
        self.validator.assert_valid(
            method_name="isApprovedForAll",
            parameter_name="_owner",
            argument_value=_owner,
        )
        _owner = self.validate_and_checksum_address(_owner)
        self.validator.assert_valid(
            method_name="isApprovedForAll",
            parameter_name="_operator",
            argument_value=_operator,
        )
        _operator = self.validate_and_checksum_address(_operator)
        return (_owner, _operator)

    def call(
        self, _owner: str, _operator: str, tx_params: Optional[TxParams] = None
    ) -> bool:
        """Execute underlying contract method via eth_call.

        :param _operator: The address that acts on behalf of the owner
        :param _owner: The address that owns the NFTs
        :param tx_params: transaction parameters
        :returns: True if `_operator` is an approved operator for `_owner`,
            false otherwise
        """
        (_owner, _operator) = self.validate_and_normalize_inputs(
            _owner, _operator
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner, _operator).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self, _owner: str, _operator: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param _operator: The address that acts on behalf of the owner
        :param _owner: The address that owns the NFTs
        :param tx_params: transaction parameters
        :returns: True if `_operator` is an approved operator for `_owner`,
            false otherwise
        """
        (_owner, _operator) = self.validate_and_normalize_inputs(
            _owner, _operator
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner, _operator).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, _owner: str, _operator: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (_owner, _operator) = self.validate_and_normalize_inputs(
            _owner, _operator
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner, _operator).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, _owner: str, _operator: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (_owner, _operator) = self.validate_and_normalize_inputs(
            _owner, _operator
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_owner, _operator).buildTransaction(
            tx_params.as_dict()
        )


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class ERC721Token:
    """Wrapper class for ERC721Token Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """

    get_approved: GetApprovedMethod
    """Constructor-initialized instance of
    :class:`GetApprovedMethod`.
    """

    approve: ApproveMethod
    """Constructor-initialized instance of
    :class:`ApproveMethod`.
    """

    transfer_from: TransferFromMethod
    """Constructor-initialized instance of
    :class:`TransferFromMethod`.
    """

    safe_transfer_from1: SafeTransferFrom1Method
    """Constructor-initialized instance of
    :class:`SafeTransferFrom1Method`.
    """

    owner_of: OwnerOfMethod
    """Constructor-initialized instance of
    :class:`OwnerOfMethod`.
    """

    balance_of: BalanceOfMethod
    """Constructor-initialized instance of
    :class:`BalanceOfMethod`.
    """

    set_approval_for_all: SetApprovalForAllMethod
    """Constructor-initialized instance of
    :class:`SetApprovalForAllMethod`.
    """

    safe_transfer_from2: SafeTransferFrom2Method
    """Constructor-initialized instance of
    :class:`SafeTransferFrom2Method`.
    """

    is_approved_for_all: IsApprovedForAllMethod
    """Constructor-initialized instance of
    :class:`IsApprovedForAllMethod`.
    """

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        validator: ERC721TokenValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param provider: instance of :class:`web3.providers.base.BaseProvider`
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        self.contract_address = contract_address

        if not validator:
            validator = ERC721TokenValidator(provider, contract_address)

        self._web3_eth = Web3(  # type: ignore # pylint: disable=no-member
            provider
        ).eth

        functions = self._web3_eth.contract(
            address=to_checksum_address(contract_address),
            abi=ERC721Token.abi(),
        ).functions

        self.get_approved = GetApprovedMethod(
            provider, contract_address, functions.getApproved, validator
        )

        self.approve = ApproveMethod(
            provider, contract_address, functions.approve, validator
        )

        self.transfer_from = TransferFromMethod(
            provider, contract_address, functions.transferFrom, validator
        )

        self.safe_transfer_from1 = SafeTransferFrom1Method(
            provider, contract_address, functions.safeTransferFrom, validator
        )

        self.owner_of = OwnerOfMethod(
            provider, contract_address, functions.ownerOf, validator
        )

        self.balance_of = BalanceOfMethod(
            provider, contract_address, functions.balanceOf, validator
        )

        self.set_approval_for_all = SetApprovalForAllMethod(
            provider, contract_address, functions.setApprovalForAll, validator
        )

        self.safe_transfer_from2 = SafeTransferFrom2Method(
            provider, contract_address, functions.safeTransferFrom, validator
        )

        self.is_approved_for_all = IsApprovedForAllMethod(
            provider, contract_address, functions.isApprovedForAll, validator
        )

    def get_transfer_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Transfer event.

        :param tx_hash: hash of transaction emitting Transfer event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=ERC721Token.abi(),
            )
            .events.Transfer()
            .processReceipt(tx_receipt)
        )

    def get_approval_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Approval event.

        :param tx_hash: hash of transaction emitting Approval event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=ERC721Token.abi(),
            )
            .events.Approval()
            .processReceipt(tx_receipt)
        )

    def get_approval_for_all_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for ApprovalForAll event.

        :param tx_hash: hash of transaction emitting ApprovalForAll event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=ERC721Token.abi(),
            )
            .events.ApprovalForAll()
            .processReceipt(tx_receipt)
        )

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"constant":true,"inputs":[{"name":"_tokenId","type":"uint256"}],"name":"getApproved","outputs":[{"name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_approved","type":"address"},{"name":"_tokenId","type":"uint256"}],"name":"approve","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_tokenId","type":"uint256"}],"name":"transferFrom","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_tokenId","type":"uint256"}],"name":"safeTransferFrom","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"name":"_tokenId","type":"uint256"}],"name":"ownerOf","outputs":[{"name":"","type":"address"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"}],"name":"balanceOf","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"_operator","type":"address"},{"name":"_approved","type":"bool"}],"name":"setApprovalForAll","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_tokenId","type":"uint256"},{"name":"_data","type":"bytes"}],"name":"safeTransferFrom","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"},{"name":"_operator","type":"address"}],"name":"isApprovedForAll","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"view","type":"function"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":true,"name":"_tokenId","type":"uint256"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_approved","type":"address"},{"indexed":true,"name":"_tokenId","type":"uint256"}],"name":"Approval","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_operator","type":"address"},{"indexed":false,"name":"_approved","type":"bool"}],"name":"ApprovalForAll","type":"event"}]'  # noqa: E501 (line-too-long)
        )


# pylint: disable=too-many-lines



"""Tests for ERC20Token wrapper."""

from decimal import Decimal

import pytest

from zero_ex.contract_addresses import NETWORK_TO_ADDRESSES, NetworkId
from zero_ex.contract_wrappers import TxParams
from zero_ex.contract_wrappers.erc20_token import ERC20Token


MAX_ALLOWANCE = int("{:.0f}".format(Decimal(2) ** 256 - 1))


@pytest.fixture(scope="module")
def erc20_wrapper(ganache_provider):
    """Get an instance of ERC20Token wrapper class for testing."""
    return ERC20Token(
        ganache_provider, NETWORK_TO_ADDRESSES[NetworkId.GANACHE].ether_token
    )


def test_erc20_wrapper__balance_of(
    accounts,
    erc20_wrapper,  # pylint: disable=redefined-outer-name
    weth_instance,  # pylint: disable=redefined-outer-name
):
    """Test getting baance of an account for an ERC20 token."""
    acc1_original_weth_balance = erc20_wrapper.balance_of.call(accounts[0])
    acc2_original_weth_balance = erc20_wrapper.balance_of.call(accounts[1])

    expected_difference = 1 * 10 ** 18

    weth_instance.functions.deposit().transact(
        {"from": accounts[0], "value": expected_difference}
    )
    weth_instance.functions.deposit().transact(
        {"from": accounts[1], "value": expected_difference}
    )
    acc1_weth_balance = erc20_wrapper.balance_of.call(accounts[0])
    acc2_weth_balance = erc20_wrapper.balance_of.call(accounts[1])

    assert (
        acc1_weth_balance - acc1_original_weth_balance == expected_difference
    )
    assert (
        acc2_weth_balance - acc2_original_weth_balance == expected_difference
    )


def test_erc20_wrapper__approve(
    accounts,
    erc20_proxy_address,
    erc20_wrapper,  # pylint: disable=redefined-outer-name
):
    """Test approving one account to spend balance from another account."""
    erc20_wrapper.approve.send_transaction(
        erc20_proxy_address,
        MAX_ALLOWANCE,
        tx_params=TxParams(from_=accounts[0]),
    )
    erc20_wrapper.approve.send_transaction(
        erc20_proxy_address,
        MAX_ALLOWANCE,
        tx_params=TxParams(from_=accounts[1]),
    )

    acc_1_weth_allowance = erc20_wrapper.allowance.call(
        accounts[0], erc20_proxy_address
    )
    acc_2_weth_allowance = erc20_wrapper.allowance.call(
        accounts[1], erc20_proxy_address
    )

    assert acc_1_weth_allowance == MAX_ALLOWANCE
    assert acc_2_weth_allowance == MAX_ALLOWANCE



"""Generated wrapper for WETH9 Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for WETH9 below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        WETH9Validator,
    )
except ImportError:

    class WETH9Validator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


class NameMethod(ContractMethod):
    """Various interfaces to the name method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())


class ApproveMethod(ContractMethod):
    """Various interfaces to the approve method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, guy: str, wad: int):
        """Validate the inputs to the approve method."""
        self.validator.assert_valid(
            method_name="approve", parameter_name="guy", argument_value=guy,
        )
        guy = self.validate_and_checksum_address(guy)
        self.validator.assert_valid(
            method_name="approve", parameter_name="wad", argument_value=wad,
        )
        # safeguard against fractional inputs
        wad = int(wad)
        return (guy, wad)

    def call(
        self, guy: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[bool, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (guy, wad) = self.validate_and_normalize_inputs(guy, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(guy, wad).call(tx_params.as_dict())

    def send_transaction(
        self, guy: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (guy, wad) = self.validate_and_normalize_inputs(guy, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(guy, wad).transact(tx_params.as_dict())

    def estimate_gas(
        self, guy: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (guy, wad) = self.validate_and_normalize_inputs(guy, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(guy, wad).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, guy: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (guy, wad) = self.validate_and_normalize_inputs(guy, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(guy, wad).buildTransaction(
            tx_params.as_dict()
        )


class TotalSupplyMethod(ContractMethod):
    """Various interfaces to the totalSupply method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())


class TransferFromMethod(ContractMethod):
    """Various interfaces to the transferFrom method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, src: str, dst: str, wad: int):
        """Validate the inputs to the transferFrom method."""
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="src",
            argument_value=src,
        )
        src = self.validate_and_checksum_address(src)
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="dst",
            argument_value=dst,
        )
        dst = self.validate_and_checksum_address(dst)
        self.validator.assert_valid(
            method_name="transferFrom",
            parameter_name="wad",
            argument_value=wad,
        )
        # safeguard against fractional inputs
        wad = int(wad)
        return (src, dst, wad)

    def call(
        self,
        src: str,
        dst: str,
        wad: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[bool, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (src, dst, wad) = self.validate_and_normalize_inputs(src, dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(src, dst, wad).call(tx_params.as_dict())

    def send_transaction(
        self,
        src: str,
        dst: str,
        wad: int,
        tx_params: Optional[TxParams] = None,
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (src, dst, wad) = self.validate_and_normalize_inputs(src, dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(src, dst, wad).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self,
        src: str,
        dst: str,
        wad: int,
        tx_params: Optional[TxParams] = None,
    ) -> int:
        """Estimate gas consumption of method call."""
        (src, dst, wad) = self.validate_and_normalize_inputs(src, dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(src, dst, wad).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self,
        src: str,
        dst: str,
        wad: int,
        tx_params: Optional[TxParams] = None,
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (src, dst, wad) = self.validate_and_normalize_inputs(src, dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(src, dst, wad).buildTransaction(
            tx_params.as_dict()
        )


class WithdrawMethod(ContractMethod):
    """Various interfaces to the withdraw method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, wad: int):
        """Validate the inputs to the withdraw method."""
        self.validator.assert_valid(
            method_name="withdraw", parameter_name="wad", argument_value=wad,
        )
        # safeguard against fractional inputs
        wad = int(wad)
        return wad

    def call(
        self, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (wad) = self.validate_and_normalize_inputs(wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(wad).call(tx_params.as_dict())

    def send_transaction(
        self, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (wad) = self.validate_and_normalize_inputs(wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(wad).transact(tx_params.as_dict())

    def estimate_gas(
        self, wad: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (wad) = self.validate_and_normalize_inputs(wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(wad).estimateGas(tx_params.as_dict())

    def build_transaction(
        self, wad: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (wad) = self.validate_and_normalize_inputs(wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(wad).buildTransaction(
            tx_params.as_dict()
        )


class DecimalsMethod(ContractMethod):
    """Various interfaces to the decimals method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())


class BalanceOfMethod(ContractMethod):
    """Various interfaces to the balanceOf method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: str):
        """Validate the inputs to the balanceOf method."""
        self.validator.assert_valid(
            method_name="balanceOf",
            parameter_name="index_0",
            argument_value=index_0,
        )
        index_0 = self.validate_and_checksum_address(index_0)
        return index_0

    def call(self, index_0: str, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).call(tx_params.as_dict())

    def send_transaction(
        self, index_0: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).transact(tx_params.as_dict())

    def estimate_gas(
        self, index_0: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, index_0: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0) = self.validate_and_normalize_inputs(index_0)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0).buildTransaction(
            tx_params.as_dict()
        )


class SymbolMethod(ContractMethod):
    """Various interfaces to the symbol method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())


class TransferMethod(ContractMethod):
    """Various interfaces to the transfer method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, dst: str, wad: int):
        """Validate the inputs to the transfer method."""
        self.validator.assert_valid(
            method_name="transfer", parameter_name="dst", argument_value=dst,
        )
        dst = self.validate_and_checksum_address(dst)
        self.validator.assert_valid(
            method_name="transfer", parameter_name="wad", argument_value=wad,
        )
        # safeguard against fractional inputs
        wad = int(wad)
        return (dst, wad)

    def call(
        self, dst: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[bool, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (dst, wad) = self.validate_and_normalize_inputs(dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(dst, wad).call(tx_params.as_dict())

    def send_transaction(
        self, dst: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (dst, wad) = self.validate_and_normalize_inputs(dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(dst, wad).transact(tx_params.as_dict())

    def estimate_gas(
        self, dst: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (dst, wad) = self.validate_and_normalize_inputs(dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(dst, wad).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, dst: str, wad: int, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (dst, wad) = self.validate_and_normalize_inputs(dst, wad)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(dst, wad).buildTransaction(
            tx_params.as_dict()
        )


class DepositMethod(ContractMethod):
    """Various interfaces to the deposit method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def call(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[None, Union[HexBytes, bytes]]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().call(tx_params.as_dict())

    def send_transaction(
        self, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())


class AllowanceMethod(ContractMethod):
    """Various interfaces to the allowance method."""

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        contract_function: ContractFunction,
        validator: Validator = None,
    ):
        """Persist instance data."""
        super().__init__(provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, index_0: str, index_1: str):
        """Validate the inputs to the allowance method."""
        self.validator.assert_valid(
            method_name="allowance",
            parameter_name="index_0",
            argument_value=index_0,
        )
        index_0 = self.validate_and_checksum_address(index_0)
        self.validator.assert_valid(
            method_name="allowance",
            parameter_name="index_1",
            argument_value=index_1,
        )
        index_1 = self.validate_and_checksum_address(index_1)
        return (index_0, index_1)

    def call(
        self, index_0: str, index_1: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters

        """
        (index_0, index_1) = self.validate_and_normalize_inputs(
            index_0, index_1
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).call(
            tx_params.as_dict()
        )

    def send_transaction(
        self, index_0: str, index_1: str, tx_params: Optional[TxParams] = None
    ) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters

        """
        (index_0, index_1) = self.validate_and_normalize_inputs(
            index_0, index_1
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).transact(
            tx_params.as_dict()
        )

    def estimate_gas(
        self, index_0: str, index_1: str, tx_params: Optional[TxParams] = None
    ) -> int:
        """Estimate gas consumption of method call."""
        (index_0, index_1) = self.validate_and_normalize_inputs(
            index_0, index_1
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).estimateGas(
            tx_params.as_dict()
        )

    def build_transaction(
        self, index_0: str, index_1: str, tx_params: Optional[TxParams] = None
    ) -> dict:
        """Construct calldata to be used as input to the method."""
        (index_0, index_1) = self.validate_and_normalize_inputs(
            index_0, index_1
        )
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(index_0, index_1).buildTransaction(
            tx_params.as_dict()
        )


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class WETH9:
    """Wrapper class for WETH9 Solidity contract."""

    name: NameMethod
    """Constructor-initialized instance of
    :class:`NameMethod`.
    """

    approve: ApproveMethod
    """Constructor-initialized instance of
    :class:`ApproveMethod`.
    """

    total_supply: TotalSupplyMethod
    """Constructor-initialized instance of
    :class:`TotalSupplyMethod`.
    """

    transfer_from: TransferFromMethod
    """Constructor-initialized instance of
    :class:`TransferFromMethod`.
    """

    withdraw: WithdrawMethod
    """Constructor-initialized instance of
    :class:`WithdrawMethod`.
    """

    decimals: DecimalsMethod
    """Constructor-initialized instance of
    :class:`DecimalsMethod`.
    """

    balance_of: BalanceOfMethod
    """Constructor-initialized instance of
    :class:`BalanceOfMethod`.
    """

    symbol: SymbolMethod
    """Constructor-initialized instance of
    :class:`SymbolMethod`.
    """

    transfer: TransferMethod
    """Constructor-initialized instance of
    :class:`TransferMethod`.
    """

    deposit: DepositMethod
    """Constructor-initialized instance of
    :class:`DepositMethod`.
    """

    allowance: AllowanceMethod
    """Constructor-initialized instance of
    :class:`AllowanceMethod`.
    """

    def __init__(
        self,
        provider: BaseProvider,
        contract_address: str,
        validator: WETH9Validator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param provider: instance of :class:`web3.providers.base.BaseProvider`
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        self.contract_address = contract_address

        if not validator:
            validator = WETH9Validator(provider, contract_address)

        self._web3_eth = Web3(  # type: ignore # pylint: disable=no-member
            provider
        ).eth

        functions = self._web3_eth.contract(
            address=to_checksum_address(contract_address), abi=WETH9.abi()
        ).functions

        self.name = NameMethod(
            provider, contract_address, functions.name, validator
        )

        self.approve = ApproveMethod(
            provider, contract_address, functions.approve, validator
        )

        self.total_supply = TotalSupplyMethod(
            provider, contract_address, functions.totalSupply, validator
        )

        self.transfer_from = TransferFromMethod(
            provider, contract_address, functions.transferFrom, validator
        )

        self.withdraw = WithdrawMethod(
            provider, contract_address, functions.withdraw, validator
        )

        self.decimals = DecimalsMethod(
            provider, contract_address, functions.decimals, validator
        )

        self.balance_of = BalanceOfMethod(
            provider, contract_address, functions.balanceOf, validator
        )

        self.symbol = SymbolMethod(
            provider, contract_address, functions.symbol, validator
        )

        self.transfer = TransferMethod(
            provider, contract_address, functions.transfer, validator
        )

        self.deposit = DepositMethod(
            provider, contract_address, functions.deposit, validator
        )

        self.allowance = AllowanceMethod(
            provider, contract_address, functions.allowance, validator
        )

    def get_approval_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Approval event.

        :param tx_hash: hash of transaction emitting Approval event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=WETH9.abi(),
            )
            .events.Approval()
            .processReceipt(tx_receipt)
        )

    def get_transfer_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Transfer event.

        :param tx_hash: hash of transaction emitting Transfer event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=WETH9.abi(),
            )
            .events.Transfer()
            .processReceipt(tx_receipt)
        )

    def get_deposit_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Deposit event.

        :param tx_hash: hash of transaction emitting Deposit event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=WETH9.abi(),
            )
            .events.Deposit()
            .processReceipt(tx_receipt)
        )

    def get_withdrawal_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for Withdrawal event.

        :param tx_hash: hash of transaction emitting Withdrawal event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return (
            self._web3_eth.contract(
                address=to_checksum_address(self.contract_address),
                abi=WETH9.abi(),
            )
            .events.Withdrawal()
            .processReceipt(tx_receipt)
        )

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"guy","type":"address"},{"name":"wad","type":"uint256"}],"name":"approve","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"src","type":"address"},{"name":"dst","type":"address"},{"name":"wad","type":"uint256"}],"name":"transferFrom","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[{"name":"wad","type":"uint256"}],"name":"withdraw","outputs":[],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint8"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{"name":"index_0","type":"address"}],"name":"balanceOf","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[],"name":"symbol","outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{"name":"dst","type":"address"},{"name":"wad","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":false,"inputs":[],"name":"deposit","outputs":[],"payable":true,"stateMutability":"payable","type":"function"},{"constant":true,"inputs":[{"name":"index_0","type":"address"},{"name":"index_1","type":"address"}],"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},{"payable":true,"stateMutability":"payable","type":"fallback"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_spender","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Approval","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},{"indexed":true,"name":"_to","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Deposit","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Withdrawal","type":"event"}]'  # noqa: E501 (line-too-long)
        )


# pylint: disable=too-many-lines
